#include "sprox_desfire_i.h"

#include <stdlib.h>

/*
 * DESFire cryptographic functions
 * -------------------------------
 */

/* Compute a DES MAC in CBC receive mode */
static void DesFire_DES_MAC(BYTE * data, WORD blocks, const BYTE * key, BYTE * result)
{
  WORD    i;
  BYTE    j;
  BYTE    carry[8], buffer[8];

  SPROX_DES_Init(NULL, (BYTE *) key);
  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    memcpy(buffer, &data[8 * i], 8);
    for (j = 0; j < 8; j++)
      buffer[j] ^= carry[j];    /* P  <- P XOR IV  */
    SPROX_DES_Encrypt(NULL, carry, buffer); /* IV <- DES(P)    */
  }
  memcpy(result, carry, 4);     /* This is the MAC */
}

/* Compute a 3DES MAC in CBC receive mode */
static void DesFire_3DES_MAC(BYTE * data, WORD blocks, const BYTE * key, BYTE * result)
{
  WORD    i;
  BYTE    j;
  BYTE    carry[8], buffer[8];

  SPROX_3DES_Init(NULL, (BYTE *) & key[0], (BYTE *) & key[8], NULL);
  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    memcpy(buffer, &data[8 * i], 8);
    for (j = 0; j < 8; j++)
      buffer[j] ^= carry[j];    /* P  <- P XOR IV  */
    SPROX_3DES_Encrypt(NULL, carry, buffer);  /* IV <- 3DES(P)   */
  }
  memcpy(result, carry, 4);     /* This is the MAC */
}

/* DES decryption in CBC send mode */
static void DesFire_DES_Send(BYTE * data, WORD blocks, const BYTE * key)
{
  WORD    i;
  BYTE    j;
  BYTE    carry[8];

  SPROX_DES_Init(NULL, (BYTE *) key);
  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
    SPROX_DES_Decrypt(NULL, carry, &data[8 * i]); /* IV <- iDES(P)   */
    memcpy(&data[8 * i], carry, 8); /* P  <- IV        */
  }
}

/* 3DES decryption in CBC send mode */
static void DesFire_3DES_Send(BYTE * data, WORD blocks, const BYTE * key)
{
  DWORD   i;
  BYTE    j;
  BYTE    carry[8];

  SPROX_3DES_Init(NULL, (BYTE *) & key[0], (BYTE *) & key[8], NULL);
  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
    SPROX_3DES_Decrypt(NULL, carry, &data[8 * i]);  /* IV <- i3DES(P)   */
    memcpy(&data[8 * i], carry, 8); /* P  <- IV        */
  }
}

/* DES decryption in CBC receive mode */
static void DesFire_DES_Recv(BYTE * data, WORD blocks, const BYTE * key)
{
  WORD    i;
  BYTE    j;
  BYTE    carry[8], buffer[8];

  SPROX_DES_Init(NULL, (BYTE *) key);
  memset(buffer, 0x00, 8);      /* B  <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    memcpy(carry, buffer, 8);   /* IV <- B         */
    memcpy(buffer, &data[8 * i], 8);  /* B  <- P         */
    SPROX_DES_Decrypt(NULL, &data[8 * i], buffer);  /* P  <- iDES(B)   */
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
  }
}

/* 3DES decryption in CBC receive mode */
static void DesFire_3DES_Recv(BYTE * data, WORD blocks, const BYTE * key)
{
  DWORD   i;
  BYTE    j;
  BYTE    carry[8], buffer[8];

  SPROX_3DES_Init(NULL, (BYTE *) & key[0], (BYTE *) & key[8], NULL);
  memset(buffer, 0x00, 8);      /* B  <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    memcpy(carry, buffer, 8);   /* IV <- B         */
    memcpy(buffer, &data[8 * i], 8);  /* B  <- P         */
    SPROX_3DES_Decrypt(NULL, &data[8 * i], buffer); /* P  <- iDES(B)   */
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
  }
}

/*
 * Compute the MAC for DESFire MACED communication
 * -----------------------------------------------
 */
void DesFire_xDES_MAC(BYTE * data, WORD blocks, const BYTE * key, BYTE * result)
{
  BYTE    mac32[4];

  /* Well formatted buffer */
  if (memcmp(&key[0], &key[8], 8))
  {
    /* 3DES */
    DesFire_3DES_MAC(data, blocks, key, mac32);
  } else
  {
    /* DES */
    DesFire_DES_MAC(data, blocks, key, mac32);
  }

  if (result != NULL)
    memcpy(result, mac32, 4);
}

void DesFire_xDES_MAC_b(BYTE * data, DWORD size, const BYTE * key, BYTE * result)
{
  if (size % 8)
  {
    WORD    blocks;
    BYTE   *buffer;

    blocks = 1 + (WORD) (size / 8);
    buffer = calloc(8, blocks);
    if (buffer == NULL)
      return;

    memcpy(buffer, data, size);
    DesFire_xDES_MAC(buffer, blocks, key, result);

    free(buffer);
  } else
  {
    DesFire_xDES_MAC(data, (WORD) (size / 8), key, result);
  }
}

/*
 * Decipher data received from the DESFire
 * ---------------------------------------
 */
void DesFire_xDES_Recv(BYTE * data, WORD blocks, const BYTE * key)
{
  if (memcmp(&key[0], &key[8], 8))
  {
    /* 3DES */
    DesFire_3DES_Recv(data, blocks, key);
  } else
  {
    /* DES */
    DesFire_DES_Recv(data, blocks, key);
  }
}

/*
 * Encipher data to be sent to the DESFire
 * ---------------------------------------
 */
void DesFire_xDES_Send(BYTE * data, WORD blocks, const BYTE * key)
{
  if (memcmp(&key[0], &key[8], 8))
  {
    /* 3DES */
    DesFire_3DES_Send(data, blocks, key);
  } else
  {
    /* DES */
    DesFire_DES_Send(data, blocks, key);
  }
}
