#include "cardappz.h"
#include "hardware/springprox/springprox.h"

#include <assert.h>
#include <stdio.h>

static int JaycosTransmitEx(CARDAPPZ_CTX_ST * ctx, unsigned char cmd[],
                            int cmdlen, unsigned char rsp[], int *rsplen)
{
  SWORD rc;

  assert(ctx != NULL);

  if (ctx->trace >= DUMPMORE)
  {
    int i;

    printf("Card<");
    for (i = 0; i < cmdlen; i++)
      printf(" %02X", cmd[i]);
    printf("\n");
  }

  /* T=CL transmit using SpringProx reader  */
  /* not difficult to change for PC/SC mode */
  if (rsplen != NULL)
  {
    WORD t = (WORD) * rsplen;

    rc = SPROX_TclB_Exchange(0xFF, 1, 0xFF, cmd, (WORD) cmdlen, rsp, &t);
    *rsplen = t;
  } else
  {
    rc = SPROX_TclB_Exchange(0xFF, 1, 0xFF, cmd, (WORD) cmdlen, rsp, NULL);
  }

  if (ctx->trace >= DUMPMORE)
  {
    int i;

    printf("Card>");
    if ((rsp != NULL) && (rsplen != NULL))
      for (i = 0; i < *rsplen; i++)
        printf(" %02X", rsp[i]);
    printf("\n");
  }

  return rc;
}

static int JaycosTransmit(CARDAPPZ_CTX_ST * ctx, unsigned char CLA,
                          unsigned char INS, unsigned char P1,
                          unsigned char P2, unsigned char data[], int LC,
                          int LE, unsigned char rsp[], int *rsplen,
                          unsigned short *status)
{
  unsigned char io_buffer[128];
  int o_len, i_len;
  unsigned char sw[2];
  int rc;

  if ((LC + 6) >= sizeof(io_buffer))
    return CARDAPPZ_BUFFER_OVERFLOW;

  /* Build the ISO 7816-4 APDU */
  io_buffer[0] = CLA;
  io_buffer[1] = INS;
  io_buffer[2] = P1;
  io_buffer[3] = P2;

  o_len = 4;
  if (data != NULL)
  {
    io_buffer[4] = LC;
    memcpy(&io_buffer[5], data, LC);
    o_len = 5 + LC;
  }
  if (LE >= 0)
  {
    io_buffer[o_len] = (unsigned char) LE;
    o_len++;
  }

  i_len = sizeof(io_buffer);
  if (rsplen != NULL)
    if (i_len > (*rsplen + 1))
      i_len = (*rsplen + 1);

  rc = JaycosTransmitEx(ctx, io_buffer, o_len, io_buffer, &i_len);
  if (rc)
    return rc;

  if (i_len < 2)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;

  if (rsp != NULL)
  {
    memcpy(rsp, io_buffer, i_len);
    if (rsplen != NULL)
      *rsplen = i_len;
  }

  sw[0] = io_buffer[i_len - 2];
  sw[1] = io_buffer[i_len - 1];

  if (ctx->trace >= TRACE)
    if ((sw[0] != 0x90) || (sw[1] != 0x00))
      printf("Jaycos: card status %02X%02X\n", sw[0], sw[1]);

  if (status != NULL)
  {
    *status = sw[0] << 8;
    *status |= sw[1];
  }

  return CARDAPPZ_OK;
}

/*
 * JayCOS Create File (user guide AE, chapter 3.1.8)
 * -------------------------------------------------
 */
static int JaycosCreateFile(CARDAPPZ_CTX_ST * ctx, unsigned short FileID,
                            unsigned char Descriptor,
                            unsigned char Extension,
                            unsigned char AccessConditions[3],
                            unsigned char AccessKeys[3],
                            unsigned char SizeH_RecCount,
                            unsigned char SizeL_RecSize)
{
  unsigned char cmddata[10];
  int rc;
  unsigned short status;

  cmddata[0] = Descriptor;
  cmddata[1] = Extension;
  if (AccessConditions != NULL)
  {
    cmddata[2] = AccessConditions[0];
    cmddata[3] = AccessConditions[1];
    cmddata[4] = AccessConditions[2];
  } else
  {
    /* Always */
    cmddata[2] = 0xFF;
    cmddata[3] = 0xFF;
    cmddata[4] = 0xFF;
  }
  if (AccessKeys != NULL)
  {
    cmddata[5] = AccessKeys[0];
    cmddata[6] = AccessKeys[1];
    cmddata[7] = AccessKeys[2];
  } else
  {
    /* Always */
    cmddata[5] = 0xFF;
    cmddata[6] = 0xFF;
    cmddata[7] = 0xFF;
  }
  cmddata[8] = SizeH_RecCount;
  cmddata[8] = SizeL_RecSize;

  rc =
    JaycosTransmit(ctx, 0x80, 0xE0, (unsigned char) (FileID >> 8),
                   (unsigned char) FileID, cmddata, sizeof(cmddata), -1,
                   NULL, NULL, &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;             /* Wrong status word */

  return CARDAPPZ_OK;
}

/*
 * JayCOS Get Challenge (user guide AE, chapter 3.2.1)
 * ---------------------------------------------------
 */
int JaycosGetChallenge(CARDAPPZ_CTX_ST * ctx, unsigned char Challenge[8])
{
  unsigned char rsp[10];
  unsigned short status;
  int rsplen;
  int rc;

  rsplen = sizeof(rsp);
  rc =
    JaycosTransmit(ctx, 0x00, 0x84, 0x00, 0x00, NULL, 0, 8, rsp, &rsplen,
                   &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  if (Challenge != NULL)
    memcpy(Challenge, rsp, 8);

  return CARDAPPZ_OK;
}

/*
 * JayCOS External Authenticate (user guide AE, chapter 3.2.2)
 * -----------------------------------------------------------
 */
int JaycosExternalAuthenticate(CARDAPPZ_CTX_ST * ctx, unsigned char KeyIndex,
                               unsigned char KeyValue[16])
{
  unsigned char challenge[8];
  unsigned short status;
  int rc;
  TDES_CTX_ST tdes_ctx;

  TDES_Init(&tdes_ctx, &KeyValue[0], &KeyValue[8], &KeyValue[0]);

  rc = JaycosGetChallenge(ctx, challenge);
  if (rc)
    return rc;

  TDES_Encrypt(&tdes_ctx, challenge);

  rc =
    JaycosTransmit(ctx, 0x00, 0x82, 0x00, KeyIndex, challenge, 8, -1, NULL,
                   NULL, &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  return CARDAPPZ_OK;
}

/*
 * JayCOS Mutual Authenticate (user guide AE, chapter 3.2.4)
 * ---------------------------------------------------------
 */
int JaycosMutualAuthenticate(CARDAPPZ_CTX_ST * ctx, unsigned char KeyIndex,
                             unsigned char KeyValue[16])
{
  unsigned char card_challenge[8];
  unsigned char host_challenge[8];
  unsigned char buf[16];
  unsigned char rsp[10];
  int rsplen = sizeof(rsp);
  unsigned short status;
  int rc;

  rc = JaycosGetChallenge(ctx, card_challenge);
  if (rc)
    return rc;

  // TODO : encrypt card_challenge with key
  // TODO : encrypt host_challenge with key

  memcpy(&buf[0], card_challenge, 8);
  memcpy(&buf[8], host_challenge, 8);

  rc =
    JaycosTransmit(ctx, 0x00, 0x82, 0x00, KeyIndex, buf, 16, 8, rsp, &rsplen,
                   &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  // TODO : decrypt card responses and check agains host_challenge

  return CARDAPPZ_OK;
}

/*
 * JayCOS Create Master File (user guide AE, chapter 3.1.7)
 * --------------------------------------------------------
 */
int JaycosCreateMF(CARDAPPZ_CTX_ST * ctx)
{
  unsigned short status;
  int rc;

  rc =
    JaycosTransmit(ctx, 0x80, 0xE0, 0x3F, 0x00, NULL, 0, -1, NULL, NULL,
                   &status);

  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  return CARDAPPZ_OK;
}

/*
 * JayCOS Create DF (user guide AE, chapter 3.1.8)
 * -----------------------------------------------
 */
int JaycosCreateDF(CARDAPPZ_CTX_ST * ctx, unsigned short DirectoryIndex)
{
  unsigned char buf[10];
  unsigned short status;
  int rc;

  buf[0] = 0x00;                /* File Type = DF */
  buf[1] = 0x00;                /* (File Descriptor) */
  buf[2] = 0xFF;                /* Access conditions : ALWAYS */
  buf[3] = 0xFF;                /* Access conditions : ALWAYS */
  buf[4] = 0xFF;                /* Access conditions : ALWAYS */
  buf[5] = 0x00;                /* No key */
  buf[6] = 0x00;                /* No key */
  buf[7] = 0x00;                /* No key */
  buf[8] = 0x00;                /* (File Size H) */
  buf[9] = 0x00;                /* (File Size L) */

  rc =
    JaycosTransmit(ctx, 0x80, 0xE0,
                   (unsigned char) (DirectoryIndex / 0x0100),
                   (unsigned char) (DirectoryIndex % 0x0100), buf, 10, -1,
                   NULL, NULL, &status);

  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  return CARDAPPZ_OK;
}

/*
 * JayCOS Create Transparent EF (user guide AE, chapter 3.1.8)
 * -----------------------------------------------------------
 */
int JaycosCreateTransparentEF(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                              unsigned short Size)
{
  unsigned char buf[10];
  unsigned short status;
  int rc;

  buf[0] = 0x20;                /* File Type = Transparent EF */
  buf[1] = 0x00;                /* (File Descriptor) */
  buf[2] = 0xFF;                /* Access conditions : ALWAYS */
  buf[3] = 0xFF;                /* Access conditions : ALWAYS */
  buf[4] = 0xFF;                /* Access conditions : ALWAYS */
  buf[5] = 0x00;                /* No key */
  buf[6] = 0x00;                /* No key */
  buf[7] = 0x00;                /* No key */
  buf[8] = (unsigned char) (Size / 0x0100); /* (File Size H) */
  buf[9] = (unsigned char) (Size % 0x0100); /* (File Size L) */

  rc =
    JaycosTransmit(ctx, 0x80, 0xE0, 0x00, FileIndex, buf, 10, -1, NULL, NULL,
                   &status);

  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  return CARDAPPZ_OK;
}

/*
 * JayCOS Select MF (user guide AE, chapter 3.1.10)
 * ------------------------------------------------
 */
int JaycosSelectMF(CARDAPPZ_CTX_ST * ctx)
{
  unsigned char buf[2];
  unsigned short status;
  int rc;

  buf[0] = 0x3F;
  buf[1] = 0x00;

  rc =
    JaycosTransmit(ctx, 0x00, 0xA4, 0x08, 0x00, buf, 2, -1, NULL, NULL,
                   &status);

  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  return CARDAPPZ_OK;
}

/*
 * JayCOS Select DF (user guide AE, chapter 3.1.10)
 * ------------------------------------------------
 */
int JaycosSelectDF(CARDAPPZ_CTX_ST * ctx, unsigned short DirectoryIndex)
{
  unsigned char buf[2];
  unsigned short status;
  int rc;

  buf[0] = (unsigned char) (DirectoryIndex / 0x0100);
  buf[1] = (unsigned char) (DirectoryIndex % 0x0100);

  rc =
    JaycosTransmit(ctx, 0x00, 0xA4, 0x08, 0x00, buf, 2, -1, NULL, NULL,
                   &status);

  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  return CARDAPPZ_OK;
}

/*
 * Jaycos Get ATR + Get Chip number (functional specification 3.1, chapter 4.7.1)
 * ----------------------------------------------------------------------------
 */
int JaycosSelectCard(CARDAPPZ_CTX_ST * ctx)
{
  /* Select Desfire ISO applet */
  BYTE rsp[24];
  unsigned short status;
  int rsplen;
  int rc;

  assert(ctx != NULL);

  if (ctx->trace >= TRACE)
    printf("Jaycos: Get ATR\n");

  rc =
    JaycosTransmit(ctx, 0x80, 0xEC, 0x00, 0x00, NULL, 0, 0x0D, rsp, &rsplen,
                   &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  if (ctx->trace >= TRACE)
    printf("Jaycos: Get Chip Number\n");

  rc =
    JaycosTransmit(ctx, 0xB0, 0xEE, 0x00, 0x00, NULL, 0, 0x08, rsp, &rsplen,
                   &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;

  return CARDAPPZ_OK;
}


int JaycosChangeKey(CARDAPPZ_CTX_ST * ctx, unsigned char KeyIndex,
                    unsigned char KeyValue[16])
{
  return CARDAPPZ_NOT_YET_IMPLEMENTED;
}

int JaycosReadBinary(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                     unsigned short Offset, unsigned short Length,
                     unsigned char Buffer[])
{
  return CARDAPPZ_NOT_YET_IMPLEMENTED;
}

int JaycosReadBinaryEx(CARDAPPZ_CTX_ST * ctx, unsigned short Offset,
                       unsigned char Length, unsigned char Buffer[])
{
  return CARDAPPZ_NOT_YET_IMPLEMENTED;
}

int JaycosUpdateBinary(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                       unsigned short Offset, unsigned short Length,
                       unsigned char Buffer[])
{
  return CARDAPPZ_NOT_YET_IMPLEMENTED;
}

int JaycosUpdateBinaryEx(CARDAPPZ_CTX_ST * ctx, unsigned short Offset,
                         unsigned char Length, unsigned char Buffer[])
{
  return CARDAPPZ_NOT_YET_IMPLEMENTED;
}


/*
 * Blank the Jaycos card (undocumented)
 * ------------------------------------
 */
static int JaycosWriteEeprom(CARDAPPZ_CTX_ST * ctx, unsigned short Offset,
                             unsigned char Size, unsigned char *Data)
{
  unsigned short status;
  int rc;
  unsigned char data[256] = { 0 };

  if (Data == NULL)
    rc =
      JaycosTransmit(ctx, 0xB0, 0xB3, (unsigned char) (Offset / 0x0100),
                     (unsigned char) (Offset % 0x0100), data, Size, -1, NULL,
                     NULL, &status);
  else
    rc =
      JaycosTransmit(ctx, 0xB0, 0xB3, (unsigned char) (Offset / 0x0100),
                     (unsigned char) (Offset % 0x0100), Data, Size, -1, NULL,
                     NULL, &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;             /* Wrong status word */

  return CARDAPPZ_OK;
}

int JaycosBlankCard(CARDAPPZ_CTX_ST * ctx)
{
  unsigned short offset;
  int rc;
  unsigned char data1[0x12] =
    { 0x2A, 0xA1, 0xFA, 0x1F, 0x8D, 0x09, 0x17, 0xF4, 0xD2, 0x2F, 0xC8, 0x71,
    0xD2, 0x6A, 0x25, 0xD2, 0xD0, 0xFF
  };
  unsigned char data2[0x01] = { 0x03 };

  rc = JaycosWriteEeprom(ctx, 0x87FB, 0x04, NULL);
  if (rc)
    return rc;
  rc = JaycosWriteEeprom(ctx, 0x87F0, 0x10, NULL);
  if (rc)
    return rc;
  rc = JaycosWriteEeprom(ctx, 0x8000, 0x04, NULL);
  if (rc)
    return rc;
  for (offset = 0x8160; offset <= 0x8240; offset += 0x20)
  {
    rc = JaycosWriteEeprom(ctx, offset, 0x20, NULL);
    if (rc)
      return rc;
  }
  rc = JaycosWriteEeprom(ctx, 0x8006, 0x02, NULL);
  if (rc)
    return rc;
  rc = JaycosWriteEeprom(ctx, 0x8008, 0x10, NULL);
  if (rc)
    return rc;
  rc = JaycosWriteEeprom(ctx, 0x87DE, 0x12, data1);
  if (rc)
    return rc;
  rc = JaycosWriteEeprom(ctx, 0x87DD, 0x01, data2);
  if (rc)
    return rc;

  return CARDAPPZ_OK;
}
