/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2005-2007 Pro-Active SAS - www.pro-active.fr
   
  Ref_Icao.c
  ----------

  This is the reference applications that shows how to read a ICAO
  passport (must comply with LDS 1.7, no BAC)
  
  JDA 14/11/2005 : initial release
  JDA 10/01/2007 : fixed a few warnings when compiling under GCC
  
  
  NOTE : A complete open-source passport reading solution can
         found at www.wzpass.net (C library + C# for .NET GUI).
         It works with CSB & SpringProx, and implements BAC.
  

*/

#include "springprox.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#ifdef WIN32
  #include <direct.h>
#endif

#ifdef __linux
  #include <sys/stat.h>
  #include <sys/types.h>
#endif

SWORD ICAO_Connect(BYTE *pbAts, BYTE *pbAtsLength);
SWORD ICAO_Disconnect(void);
BOOL ICAO_ReadPassportPlain(char *target_directory);

/* Current activated card type (ISO 14443-A or ISO 14443-B) */
BYTE icao_card_type = 0;

/* Default card CID */
BYTE icao_card_cid = 1;

/* Max baudrate in each direction */
BYTE icao_dsi_dri = 0;

/* Limit us to small frames */
BYTE icao_small = 0;

int main(int argc, char **argv)
{
  SWORD rc;
  int i;
  char s_buffer[64];  
  BYTE ats[32];
  BYTE atslen;

  char *device = NULL;
  char *target = NULL;

  printf("Pro-Active K531/SpringProx/CSB Developer's Kit\n");
  printf("\n");
  printf("ICAO LDS 1.7 demo software\n");
  printf("---------------------------\n");
  printf("www.pro-active.fr\n\n");

  for (i=1; i<argc; i++)
  {
    if (!strcmp(argv[i], "-b0")) icao_dsi_dri = 0; else
    if (!strcmp(argv[i], "-b1")) icao_dsi_dri = 1; else
    if (!strcmp(argv[i], "-b2")) icao_dsi_dri = 2; else
    if (!strcmp(argv[i], "-b3")) icao_dsi_dri = 3; else
    if (!strcmp(argv[i], "-s")) icao_small = 1; else
    if (!strcmp(argv[i], "-o") && i<argc-1)
    {
      i++;
      target = argv[i];
      printf("Target  : '%s'\n", target);
#ifdef __linux
      if ((mkdir(target, 0) < 0) && (errno != EEXIST))
#else
      if ((mkdir(target) < 0) && (errno != EEXIST))
#endif
      {
        printf("Failed to create target directory\n");
        return 0;
      }
    } else
    if (i == argc-1)
    {
      device = argv[i];
      printf("Device : '%s'\n", device);
    } else
    {
      printf("usage: ref_icao [-b0|-b1|-b2|-b3] [-o target] [device]\n");
      printf("       -b0 : set max baudrate to 106kbit/s\n");
      printf("       -b1 : set max baudrate to 212kbit/s\n");
      printf("       -b2 : set max baudrate to 424kbit/s\n");
      printf("       -b3 : set max baudrate to 848kbit/s\n");
      printf("       -s  : limit read binary to 96bytes\n");
      printf("\n");
      return 0;
    }
  }

  rc = SPROX_GetLibrary(s_buffer, sizeof(s_buffer));
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", s_buffer);
  printf("\n");

  /* Open reader */
  /* ----------- */

  rc = SPROX_ReaderOpen(device);
  if (rc != MI_OK)
  {
    printf("Reader not found\n");
    goto done;
  }

  rc = SPROX_ReaderGetDevice(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader found on %s\n", s_buffer);

  rc = SPROX_ReaderGetFirmware(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader firwmare is %s\n", s_buffer);

  /* RF field ON */
  SPROX_ControlRF(TRUE);

  /* Try to find an ISO 14443-A card */
  /* ------------------------------- */

  SPROX_SetConfig(CFG_MODE_ISO_14443_A);
  icao_card_type = CFG_MODE_ISO_14443_A;

  atslen = sizeof(ats);
  rc = ICAO_Connect(ats, &atslen);

  if (rc == MI_OK)
  {
    /* We found an ISO 14443-A tag */
    printf("ISO 14443-A tag found, ATS=");
    for (i = 0; i < atslen; i++)
      printf("%02X", ats[i]);
    printf("\n");

    /* Do the ICAO test */
    ICAO_ReadPassportPlain(target);
    ICAO_Disconnect();
    goto close;
  }

  /* Try to find an ISO 14443-B card */
  /* ------------------------------- */

  SPROX_SetConfig(CFG_MODE_ISO_14443_B);
  icao_card_type = CFG_MODE_ISO_14443_B;

  atslen = sizeof(ats);
  rc = ICAO_Connect(ats, &atslen);

  if (rc == MI_OK)
  {
    /* We found an ISO 14443-A tag */
    printf("ISO 14443-B tag found\n");

    /* Do the ICAO test */
    ICAO_ReadPassportPlain(target);
    ICAO_Disconnect();
    goto close;
  }

close:
  /* Close the reader */
  SPROX_ControlRF(FALSE);
  SPROX_ReaderClose();

done:
  /* Display last error */
  if (rc == MI_OK)
  {
    printf("Done\n");
  } else
  {
    printf("%s (%d)\n", SPROX_GetErrorMessage(rc), rc);
  }
  return 0;
}


/*
 * ICAO_Connect
 * ------------
 * Depending on current mode (ISO 14443-A or ISO 14443-B), this performs tag activation
 * sequence up to T=CL level (ISO 14443 layer 4)
 */
SWORD ICAO_Connect(BYTE * pbAts, BYTE * pbAtsLength)
{
  SWORD rc;
  BYTE ta1;

  if (icao_card_type == CFG_MODE_ISO_14443_A)
  {
    BYTE atq[2];
    BYTE sak[1];
    BYTE snr[12];
    BYTE snrlen = sizeof(snr);

    /* ISAO 14443-A activation */
    rc = SPROX_TclA_ActivateIdle(atq, snr, &snrlen, sak);
    if (rc != MI_OK)
      return rc;                /* No tag, or tag not compliant to ISO 14443-4 */

    /* T=CL card found, enter ISO 14443-4 */
    rc = SPROX_TclA_GetAts(icao_card_cid, pbAts, pbAtsLength);
    if (rc != MI_OK)
      return rc;

    /* ATR OK, select dsi and dri */
    if (*pbAtsLength > 0)
    {
      if (pbAts[0] & 0x10)
      {
        /* TA1 is provided */
        ta1 = pbAts[1];

        if ((ta1 & 0x04) && (ta1 & 0x40) && (icao_dsi_dri >= 3))
        {
          /* Go to 848kbps */
          icao_dsi_dri = 3;
        } else
        if ((ta1 & 0x02) && (ta1 & 0x20) && (icao_dsi_dri >= 2))
        {
          /* Go to 424kbps */
          icao_dsi_dri = 2;
        } else
        if ((ta1 & 0x01) && (ta1 & 0x10) && (icao_dsi_dri >= 1))
        {
          /* Go to 212kbps */
          icao_dsi_dri = 1;
        } else
          icao_dsi_dri = 0;
        
        printf("Selecting DSI=DRI=%d\n", icao_dsi_dri);

        if (icao_dsi_dri > 0)
        {
          /* Card support higher baudrates (212, 424 or 848 kbps), let's go the fatest speed */
          rc = SPROX_TclA_Pps(icao_card_cid, icao_dsi_dri, icao_dsi_dri);
          if (rc != MI_OK)
            return rc;
        }
      }
    }

    return rc;

  } else
  if (icao_card_type == CFG_MODE_ISO_14443_B)
  {
    BYTE atq[11];

    /* ISAO 14443-B activation */
    rc = SPROX_TclB_ActivateIdle(0, atq);
    if (rc != MI_OK)
      return rc;                /* No tag, or tag not compliant to ISO 14443-4 */

    /* If card support higher baudrates (212 or 424 kbps), we'll try the fastest speed */
    ta1 = atq[8];
    if ((ta1 & 0x04) && (ta1 & 0x40) && (icao_dsi_dri >= 3))
    {
      /* Go to 848kbps */
      icao_dsi_dri = 3;
    } else
    if ((ta1 & 0x02) && (ta1 & 0x20) && (icao_dsi_dri >= 2))
    {
      /* Go to 424kbps */
      icao_dsi_dri = 2;
    } else
    if ((ta1 & 0x01) && (ta1 & 0x10) && (icao_dsi_dri >= 1))
    {
      /* Go to 212kbps */
      icao_dsi_dri = 1;
    }
    
    printf("Selecting DSI=DRI=%d\n", icao_dsi_dri);
    
    /* T=CL card found, enter ISO 14443-4 */
    rc = SPROX_TclB_AttribEx(atq, icao_card_cid, icao_dsi_dri, icao_dsi_dri);
    if (rc != MI_OK)
      return rc;

    return rc;

  } else
    return MI_LIB_CALL_ERROR;
}

/*
 * ICAO_Disconnect
 * ---------------
 * This is only a wrapper around SPROX_TclA_Deselect and SPROX_TclB_Deselect
 */
SWORD ICAO_Disconnect(void)
{
  return SPROX_Tcl_Deselect(icao_card_cid);
}

/*
 * ICAO_SelectApplet
 * -----------------
 * Select the ICAO LDS applet (A0 00 00 02 47 10 01)
 */
BOOL ICAO_SelectApplet(void)
{
  BYTE pbSendBuffer[] =
    { 0x00, 0xA4, 0x04, 0x0C, 0x07, 0xA0, 0x00, 0x00, 0x02, 0x47, 0x10,
    0x01
  };
  BYTE pbRecvBuffer[258];
  WORD wRecvLength = sizeof(pbRecvBuffer);
  SWORD rc;

  /* Send the APDU to the card */
  rc = SPROX_Tcl_Exchange(icao_card_cid, pbSendBuffer, sizeof(pbSendBuffer), pbRecvBuffer, &wRecvLength);
  if (rc != 0)
  {
    printf("SPROX_Tcl_Exchange(select applet) -> %d\n", rc);
    if ((rc == MI_NOTAGERR) || (rc == TCL_TRANSMERR_NOTAG))
    {
      printf("Card removed ?\n");
      exit(EXIT_FAILURE);
    }
    return FALSE;
  }

  /* Check that we have 9000 at the end of the answer */
  if ((pbRecvBuffer[wRecvLength - 2] != 0x90)
      || (pbRecvBuffer[wRecvLength - 1] != 0x00))
  {
    printf("ICAO applet not found (select returned %02X%02X)\n",
           pbRecvBuffer[wRecvLength - 2], pbRecvBuffer[wRecvLength - 1]);
    return FALSE;
  }

  return TRUE;
}

/*
 * ICAO_SelectDG
 * -------------
 * Select a file (DG) in ICAO structure
 */
BOOL ICAO_SelectDG(BYTE dg)
{
  BYTE pbSendBuffer[] = { 0x00, 0xA4, 0x02, 0x0C, 0x02, 0x01, 0xFF };
  BYTE pbRecvBuffer[258];
  WORD wRecvLength = sizeof(pbRecvBuffer);
  SWORD rc;

  pbSendBuffer[sizeof(pbSendBuffer) - 1] = dg;

  /* Send the APDU to the card */
  rc = SPROX_Tcl_Exchange(icao_card_cid, pbSendBuffer, sizeof(pbSendBuffer), pbRecvBuffer, &wRecvLength);
  if (rc != 0)
  {
    printf("SPROX_Tcl_Exchange(select DG%d) -> error %d\n", dg, rc);
    if ((rc == MI_NOTAGERR) || (rc == TCL_TRANSMERR_NOTAG))
    {
      printf("Card removed ?\n");
      exit(EXIT_FAILURE);
    }
    return FALSE;
  }

  /* Check that we have 9000 at the end of the answer */
  if ((pbRecvBuffer[wRecvLength - 2] != 0x90)
      || (pbRecvBuffer[wRecvLength - 1] != 0x00))
  {
    printf("ICAO DG%d not available (SW=%02X%02X)\n", dg,
           pbRecvBuffer[wRecvLength - 2], pbRecvBuffer[wRecvLength - 1]);
    return FALSE;
  }

  return TRUE;
}

/*
 * ICAO_ReadBinary
 * ---------------
 * Read the content of currently selected file, starting at offset, up to want_len bytes
 * (len must be <= 256)
 * Returns the actual length that has been read
 */
BOOL ICAO_ReadBinary(BYTE *data, WORD offset, WORD want_len, WORD *recv_len)
{
  BYTE pbSendBuffer[] = { 0x00, 0xB0, 0xCC, 0xCC, 0xCC };
  BYTE pbRecvBuffer[258];
  WORD wRecvLength = sizeof(pbRecvBuffer);
  SWORD rc;

  pbSendBuffer[2] = (BYTE) (offset / 0x0100);
  pbSendBuffer[3] = (BYTE) (offset % 0x0100);
  if (want_len <= 255)
    pbSendBuffer[4] = (BYTE) want_len;
  else if (want_len == 256)
    pbSendBuffer[4] = 0x00;
  else
    return FALSE;

  /* Send the APDU to the card */
  rc = SPROX_Tcl_Exchange(icao_card_cid, pbSendBuffer, sizeof(pbSendBuffer), pbRecvBuffer, &wRecvLength);
  if (rc != MI_OK)
  {
    printf("SPROX_Tcl_Exchange(read binary offset=%d, len=%d) -> error %d\n", offset, want_len, rc);
    return FALSE;
  }
  if (wRecvLength < 2)
    return FALSE;

  *recv_len = 0;

  /* Check that we have 9000 at the end of the answer */
  if ((pbRecvBuffer[wRecvLength - 2] != 0x90)
      || (pbRecvBuffer[wRecvLength - 1] != 0x00))
  {
    printf("ICAO read file : SW=%02X%02X at offset %d\n",
           pbRecvBuffer[wRecvLength - 2], pbRecvBuffer[wRecvLength - 1],
           offset);
    return TRUE;
  }

  /* Copy the data */
  if (data != NULL)
    memcpy(data, pbRecvBuffer, wRecvLength - 2);

  /* Actual length is... */
  *recv_len = (wRecvLength - 2);

  return TRUE;
}

/*
 * ICAO_ReadPassportPlain
 * ----------------------
 * Read the whole ICAO LDS structure
 */
BOOL ICAO_ReadPassportPlain(char *target_directory)
{
  WORD offset, recv_len;
  WORD want_len = 256;
  DWORD total = 0;
  BYTE dg;
  BYTE buffer[256];
#ifndef MAX_PATH
  #define MAX_PATH 256
#endif
  char file_name[MAX_PATH];
  FILE *fp = NULL;  
  clock_t t0, t1, tr0, tr1;

  if (icao_small) want_len = 96;

  printf("Test ICAO, max recv size=%d...\n", want_len);

  t0 = clock();

  if (!ICAO_SelectApplet())
    return FALSE;

  printf("ICAO applet selected\n");

  for (dg = 1; dg <= 16; dg++)
  {
    if (ICAO_SelectDG(dg))
    {
      printf("ICAO DG%d selected\n", dg);

      if (target_directory != NULL)
      {
        sprintf(file_name, "%s/dg_%d.bin", target_directory, dg);
        fp = fopen(file_name, "wb");
      }

      tr0 = clock();
      offset = 0;
      for (;;)
      {
        if (!ICAO_ReadBinary(buffer, offset, want_len, &recv_len))
        {
          printf("ICAO : read error in DG%d\n", dg);
          return FALSE;
        }

        if (fp != NULL)
        {
          fwrite(buffer, recv_len, 1, fp); 
        }

        offset += recv_len;
        if (recv_len == 0)
          break;
      }
      tr1 = clock();

      total += offset;

      printf("DG%d : %db read in %ldms\n", dg, offset, tr1 - tr0);

      if (fp != NULL)
      {
        fclose(fp);
        fp = NULL;
      }
    }
  }

  t1 = clock();

  printf("ICAO : total %lub read in %ldms\n", (unsigned long) total, t1 - t0);
  return TRUE;
}
